/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.qpid.proton.amqp.transport;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import org.apache.qpid.proton.amqp.UnsignedInteger;
import org.junit.Test;

public class DetachTest {

    @Test
    public void testCopy() {
        Detach detach = new Detach();
        Detach copyOf = detach.copy();

        assertNull(copyOf.getError());

        detach.setError(new ErrorCondition());
        detach.setClosed(true);
        detach.setHandle(UnsignedInteger.ONE);

        copyOf = detach.copy();

        assertNotNull(copyOf.getError());
        assertEquals(detach.getClosed(), copyOf.getClosed());
        assertEquals(detach.getHandle(), copyOf.getHandle());
    }
}
