/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.qpid.proton.engine.impl;

import static org.junit.Assert.assertEquals;

import java.nio.ByteBuffer;
import java.nio.charset.StandardCharsets;

public class TransportTestHelper
{
    public static void assertByteArrayContentEquals(byte[] expectedBytes, byte[] actualBytes)
    {
        assertEquals(new String(expectedBytes), new String(actualBytes));
    }

    public static void assertByteBufferContentEquals(byte[] expectedBytes, ByteBuffer actualByteBuffer)
    {
        ByteBuffer myByteBuffer = actualByteBuffer.duplicate();
        byte[] actualBytes = new byte[myByteBuffer.remaining()];
        myByteBuffer.get(actualBytes);

        assertByteArrayContentEquals(expectedBytes, actualBytes);
    }

    public static String pourBufferToString(ByteBuffer source)
    {
        return pourBufferToString(source, source.remaining());
    }

    public static String pourBufferToString(ByteBuffer source, int sizeRequested)
    {
        byte[] buf = new byte[sizeRequested];
        int numberRead = ByteBufferUtils.pourBufferToArray(source, buf, 0, sizeRequested);
        return new String(buf, 0, numberRead, StandardCharsets.UTF_8);
    }

    public static String stringOfLength(String value, int repeat)
    {
        StringBuilder builder = new StringBuilder();
        for(int i = 0 ; i < repeat; i++)
        {
            builder.append(value);
        }
        return builder.toString();
    }
}
