/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.qpid.proton.engine.impl.ssl;

import org.apache.qpid.proton.engine.ProtonJSslPeerDetails;


public class SslPeerDetailsImpl implements ProtonJSslPeerDetails
{
    private final String _hostname;
    private final int _port;

    /**
     * Application code should use {@link org.apache.qpid.proton.engine.SslPeerDetails.Factory#create(String, int)} instead.
     */
    public SslPeerDetailsImpl(String hostname, int port)
    {
        _hostname = hostname;
        _port = port;
    }

    @Override
    public String getHostname()
    {
        return _hostname;
    }

    @Override
    public int getPort()
    {
        return _port;
    }
}
